unit RdfParser_1_1;

// RDF Parser 1.1.6
// Delphi 2010 Implementation
//
//
// LICENSE
//
// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// "http://www.mozilla.org/MPL/"
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
// the specific language governing rights and limitations under the License.
//
// The Original Code is "RdfParser_1_1.pas".
//
// The Initial Developer of the Original Code is Dieter Khler (Heidelberg,
// Germany, "http://www.philo.de/"). Portions created by the Initial Developer
// are Copyright (C) 2003-2010 Dieter Khler. All Rights Reserved.
//
// Alternatively, the contents of this file may be used under the terms of the
// GNU General Public License Version 2 or later (the "GPL"), in which case the
// provisions of the GPL are applicable instead of those above. If you wish to
// allow use of your version of this file only under the terms of the GPL, and
// not to allow others to use your version of this file under the terms of the
// MPL, indicate your decision by deleting the provisions above and replace them
// with the notice and other provisions required by the GPL. If you do not delete
// the provisions above, a recipient may use your version of this file under the
// terms of any one of the MPL or the GPL.


// HISTORY
//
// Only major modifications are listed here.
//
// 2010-05-11 Release of version 1.1.6
// 2009-02-25 Release of version 1.1.5
// 2004-12-11 Release of version 1.1.4
// 2004-06-01 Release of version 1.1.3
//            Parsing of empty nodes supported.
// 2004-02-26 Release of version 1.1.2
// 2003-10-20 Release of version 1.1.1
// 2003-04-01 Release of version 1.1.0
// 2003-03-26 Release of version 1.0.0

interface

uses
  Rdf_1_1, dkLangUtils, dkParserUtilsWin32, RdfConst,
  StrUtils, SysUtils, Classes;

type

  ERdfParserException = class(ERdfException);

  TRdf_N_Triple_Line_Token_Type = (
    NTL_TRIPLE_TOKEN,
    NTL_COMMENT_TOKEN,
    NTL_EMPTY_LINE_TOKEN,
    NTL_END_OF_TEXT_TOKEN,
    NTL_INVALID_LINE_TOKEN
  );

  TRdfInputSource = class(TUtilsUCS4Reader)
  protected
    function GetLine: Integer; virtual;
  public
    property Line: Integer read GetLine;
  end;

  TRdfNTripleLineTokenizer = class
  protected
    FInputSource: TRdfInputSource;
    function GetLine: Integer; virtual;
  public
    constructor Create(const AInputSource: TRdfInputSource); virtual;
    procedure Read(out Symbol: TRdf_N_Triple_Line_Token_Type;
                   out NextLine: string); virtual;
    property Line: Integer read GetLine;
  end;

  TRdfCustomTokenizer = class
  public
    function Read(out RdfSubject,
                      RdfPredicate,
                      RdfObject: string): Boolean; virtual; abstract;
  end;

  TRdfNTripleTokenizer = class(TRdfCustomTokenizer)
  protected
    FLineTokenizer: TRdfNTripleLineTokenizer;
    function IsWhiteSpace(S: Char): Boolean; virtual;
  public
    constructor Create(const ALineTokenizer: TRdfNTripleLineTokenizer);
    function Read(out RdfSubject,
                      RdfPredicate,
                      RdfObject: string): Boolean; override;
  end;

  TRdfCustomAnalyzer = class
  public
    function Read(out RdfSubject,
                      RdfPredicate,
                      RdfObject,
                      RdfObjectLexical,
                      RdfObjectLanguage,
                      RdfObjectDatatype: string): Boolean; virtual; abstract;
  end;

  TRdfTokenAnalyzer = class(TRdfCustomAnalyzer)
  protected
    FTokenizer: TRdfCustomTokenizer;
  public
    constructor Create(const ATokenizer: TRdfCustomTokenizer);
    function Read(out RdfSubject,
                      RdfPredicate,
                      RdfObject,
                      RdfObjectLexical,
                      RdfObjectLanguage,
                      RdfObjectDatatype: string): Boolean; override;
  end;

  TRdfNTripleToRdfGraphParser = class(TComponent)
  private
    FTarget: TRdfFilteredGraph;
    procedure SetTarget(Value: TRdfFilteredGraph);
  protected
    procedure Notification(AComponent: TComponent;
                           Operation: TOperation); override;
  public
    function ParseFile(const Filename: string;
                         var Handle: Integer): Boolean; virtual;
    function ParseStream(const Stream: TStream;
                           var Handle: Integer): Boolean; virtual;
    function ParseString(const Expression: string;
                           var Handle: Integer): Boolean; virtual;
  published
    property Target: TRdfFilteredGraph read FTarget write SetTarget;
  end;

{Helper functions}

procedure RdfParserError(Msg: string);

function NTripleStrUnescape(S: string): UTF8String;

procedure Register;


implementation

uses
  dkCodecUtilsWin32;

procedure Register;
begin
  RegisterComponents('Open RDF',[ TRdfNTripleToRdfGraphParser ]);
end;

{Helper functions}

procedure RdfParserError(Msg: string);
begin
  raise ERdfParserException.Create(Msg);
end;

function NTripleStrUnescape(S: string): UTF8String;

  function UCS4CodeToUTF8String(Code: longint): string;
  // This function was provided by Ernst van der Pols
  const
    MaxCode: array[0..5] of longint = ($7F,$7FF,$FFFF,$1FFFFF,$3FFFFFF,$7FFFFFFF);
    FirstByte: array[0..5] of Byte = (0,$C0,$E0,$F0,$F8,$FC);
  var
    I, Mo: longint;
  begin
    Result := '';
    Mo := 0; // Get number of bytes.
    while ((Code > MaxCode[Mo]) and (Mo < 5)) do Inc(Mo);
    for I := 1 to Mo do begin
      Result := Char($80 or (Code and $3F)) + Result; // Fill bytes from rear end.
      Code := Code shr 6;
    end;
    Result := Char(FirstByte[Mo] or Code) + Result;; // Fill first byte.
  end;

var
  I, L: Integer;
begin
  Result := '';
  I := 1;
  L := Length(S);
  while I <= L do begin
    if Ord(S[I]) < $80 then begin // US-ASCII ?
      if S[I] = '\' then begin
        if I = L then raise EConvertError.Create('Invalid escape sequence found.');
        Inc(I);
        case Ord(S[I]) of
          $5C: Result := Result + '\';  // '\' --> '\'
          $22: Result := Result + '"';  // '"' --> '"'
          $6E: Result := Result + #$A;  // 'n' --> LF
          $72: Result := Result + #$D;  // 'r' --> CR
          $74: Result := Result + #$9;  // 't' --> TAB
          $75: begin                    // 'uHHHH' --> 4 hex digit encoded character
            if I+4 > L then raise EConvertError.Create('Invalid escape sequence found.');
            Result := Result + UCS4CodeToUTF8String(StrToInt('$'+Copy(S,I+1,4))); // StrToInt raises EConvertError if malformed.
            Inc(I,5);
          end;
          $55: begin                    // 'UHHHHHHHH' --> 8 hex digit encoded character
            if I+8 > L then raise EConvertError.Create('Invalid escape sequence found.');
            Result := Result + UCS4CodeToUTF8String(StrToInt('$'+Copy(S,I+1,8))); // StrToInt raises EConvertError if malformed.
            Inc(I,9);
          end;
          else raise EConvertError.Create('Invalid escape sequence found.');
        end;
      end else Result := Result + S[I];
    end else raise EConvertError.Create('Non US-ASCII character found.');
    Inc(I);
  end;
end;

{ TRdfInputSource }

function TRdfInputSource.GetLine: Integer;
begin
  Result := CurrentCharInfo.Line;
end;

{ TRdfNTripleLineTokenizer }

constructor TRdfNTripleLineTokenizer.Create(const AInputSource:
  TRdfInputSource);
begin
  inherited Create;
  FInputSource := AInputSource;
end;

function TRdfNTripleLineTokenizer.GetLine: Integer;
begin
  Result := FInputSource.Line;
end;

procedure TRdfNTripleLineTokenizer.Read(
  out Symbol: TRdf_N_Triple_Line_Token_Type;
  out NextLine: string);
// Raises an EConvertError exception via the call to FInputSource.Next(),
// if one of the characters of the next line cannot be converted to a UCS4
// code point.
const
  LF: UCS4Char = $A; // Line Feed
begin
  with FInputSource do begin
    Next;
    if CurrentCharInfo.CodePoint = $9C then begin  // STRING TERMINATOR ?
      Symbol := NTL_END_OF_TEXT_TOKEN;
      NextLine := '';
      Exit;
    end else NextLine := Char(CurrentCharInfo.CodePoint);  // ###

    if CurrentCharInfo.CodePoint <> LF then begin
      repeat
        Next;
        if CurrentCharInfo.CodePoint = $9C then begin  // STRING TERMINATOR ?
          // End of string without LF --> Error
          Symbol := NTL_INVALID_LINE_TOKEN;
          NextLine := '';
          Exit;
        end;
        NextLine := NextLine + Char(CurrentCharInfo.CodePoint);  // ###
      until CurrentCharInfo.CodePoint = LF;
    end; {if ...}
  end; {with ...}

  NextLine := TrimLeft(NextLine);

  if NextLine = '' then Symbol := NTL_EMPTY_LINE_TOKEN
  else if NextLine[1] = '#' then Symbol := NTL_COMMENT_TOKEN else begin
    NextLine := TrimRight(NextLine);
    if NextLine[Length(NextLine)] <> '.' then begin
      // No '.' at the end of a triple --> Error
      Symbol := NTL_INVALID_LINE_TOKEN;
      NextLine := '';
      exit;
    end;
    NextLine := TrimRight(LeftStr(NextLine,Length(NextLine) - 1)); // Remove closing ws*'.'
    Symbol := NTL_TRIPLE_TOKEN;
  end;
end;

{ TRdfNTripleTokenizer }

constructor TRdfNTripleTokenizer.Create(const ALineTokenizer:
  TRdfNTripleLineTokenizer);
begin
  inherited Create;
  FLineTokenizer := ALineTokenizer;
end;

function TRdfNTripleTokenizer.IsWhiteSpace(S: Char): Boolean;
begin
  case Word(S) of
    $0009, $0020: Result := True;
  else
    Result := False;
  end;
end;

function TRdfNTripleTokenizer.Read(
  out RdfSubject,
      RdfPredicate,
      RdfObject: string): Boolean;
var
  ExpressionSymbol: TRdf_N_Triple_Line_Token_Type;
  Expression: string;
  I: Integer;
  WsPos: Integer;
begin
  Result := False;
  while True do begin
    FLineTokenizer.Read(ExpressionSymbol,Expression);
    case ExpressionSymbol of
      NTL_TRIPLE_TOKEN: begin
        WsPos := 0;
        for I := 1 to Length(Expression) do
          if IsWhiteSpace(Expression[I]) then begin
            WsPos := I;
            Break;
          end;
        if WsPos < 2 then RdfParserError(SRdfPredicateNotFound);
        RdfSubject := LeftStr(Expression,WsPos-1);
        Expression := TrimLeft(Copy(Expression,WsPos+1,Length(Expression)));
        WsPos := 0;
        for I := 1 to Length(Expression) do
          if IsWhiteSpace(Expression[I]) then begin
            WsPos := I;
            Break;
          end;
        if WsPos < 2 then RdfParserError(SRdfObjectNotFound);
        RdfPredicate := LeftStr(Expression,WsPos-1);
        RdfObject := TrimLeft(Copy(Expression,WsPos+1,Length(Expression)));
        if RdfObject = '' then RdfParserError(SRdfObjectNotFound);
        Result := True;
        Exit;
      end;
      NTL_END_OF_TEXT_TOKEN: exit;
      NTL_INVALID_LINE_TOKEN: RdfParserError(SRdfInvalidLine);
    end;
  end;
end;

{ TRdfTokenAnalyzer }

constructor TRdfTokenAnalyzer.Create(const ATokenizer: TRdfCustomTokenizer);
begin
  inherited Create;
  FTokenizer := ATokenizer;
end;

function TRdfTokenAnalyzer.Read(out RdfSubject, RdfPredicate, RdfObject,
  RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype: string): Boolean;
var
  SlashFound: Boolean;
  I: Integer;
  LPos: Integer;
  QPos: Integer;
begin
  Result := FTokenizer.Read(RdfSubject, RdfPredicate, RdfObject);
  if Result then begin
    if RdfObject[1] = '"' then begin
      SlashFound := False;
      QPos := 0;
      I := 2;
      repeat
        if SlashFound then begin
          SlashFound := False;
        end else begin
          if RdfObject[I] = '"' then begin
            QPos := I;
            Break;
          end else if RdfObject[I] = '\' then begin
            SlashFound := True;
          end;
        end;
        Inc(I);
      until I > Length(RdfObject);
      if QPos = 0 then RdfParserError(SRdfInvalidObject);
      RdfObjectLexical := NTripleStrUnescape(Copy(RdfObject, 2, QPos - 2));
      if QPos = Length(RdfObject) then begin
        RdfObjectLanguage := '';
        RdfObjectDatatype := '';
      end else begin
        if RdfObject[Qpos + 1] = '@' then begin
          LPos := 0;
          for I := Qpos + 2 to Length(RdfObject) do
            if RdfObject[I] = '^' then begin
              LPos := I;
              Break;
            end;
          if LPos = 0 then begin
            RdfObjectLanguage := copy(RdfObject, QPos + 2, Length(RdfObject));
            RdfObjectDatatype := '';
            RdfObject := '';
            Exit;
          end else begin
            RdfObjectLanguage := copy(RdfObject, QPos + 2, LPos - QPos - 2);
            QPos := LPos;
          end;
          if not IsRFC3066LanguageTag(RdfObjectLanguage) then
            RdfParserError(SRdfInvalidObject);
        end else if RdfObject[Qpos + 1] = '^' then begin
          RdfObjectLanguage := '';
          Inc(Qpos);
        end else RdfParserError(SRdfInvalidObject);
        if Qpos > Length(RdfObject) then RdfParserError(SRdfInvalidObject);
        if RdfObject[Qpos + 1] <> '^' then RdfParserError(SRdfInvalidObject);
        RdfObjectDatatype := copy(RdfObject, QPos + 2, Length(RdfObject));
        if RdfObjectDatatype = '' then RdfParserError(SRdfInvalidObject);
      end;
      RdfObject := '';
    end else begin
      RdfObjectLexical  := '';
      RdfObjectLanguage := '';
      RdfObjectDatatype := '';
    end;
  end else begin
    RdfSubject        := '';
    RdfPredicate      := '';
    RdfObject         := '';
    RdfObjectLexical  := '';
    RdfObjectLanguage := '';
    RdfObjectDatatype := '';
  end;
end;

{ TRdfNTripleToRdfGraphParser }

procedure TRdfNTripleToRdfGraphParser.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent = FTarget)
    then FTarget := nil;
end;

function TRdfNTripleToRdfGraphParser.ParseFile(const FileName: string;
  var Handle: Integer): Boolean;
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    Result := ParseStream(Stream, Handle);
  finally
    Stream.Free;
  end;
end;

function TRdfNTripleToRdfGraphParser.ParseStream(const Stream: TStream;
                                                   var Handle: Integer): Boolean;
var
  InputSource: TRdfInputSource;
  LineTokenizer: TRdfNTripleLineTokenizer;
  NewRdfObject: string;
  NewRdfSubject: string;
  NodeIDMap: TStringList;  // Used to store information about the mapping of
                           // names of empty nodes in a file to their equivalent
                           // names in the target RDF graph.
  RdfSubject: string;
  RdfPredicate: string;
  RdfObject: string;
  RdfObjectLexical: string;
  RdfObjectLanguage: string;
  RdfObjectDatatype: string;
  TripleAnalyzer: TRdfTokenAnalyzer;
  TripleTokenizer: TRdfNTripleTokenizer;
begin
  if not Assigned(Target) then RdfParserError(SRdfTargetNotSpecified);
  if not Target.CanModify then RdfParserError(SRdfTargetReadOnly);

  try
    TripleAnalyzer := nil;
    TripleTokenizer := nil;
    LineTokenizer := nil;
    InputSource := nil;
    NodeIDMap := nil;
    try
      InputSource := TRdfInputSource.Create(Stream, 4096, 'US-ASCII', 0, 0, 0, 0, 1);
      LineTokenizer := TRdfNTripleLineTokenizer.Create(InputSource);
      TripleTokenizer := TRdfNTripleTokenizer.Create(LineTokenizer);
      TripleAnalyzer := TRdfTokenAnalyzer.Create(TripleTokenizer);
      NodeIDMap := TStringList.Create;
      Target.DisableControls;
      try
        while TripleAnalyzer.Read(RdfSubject, RdfPredicate, RdfObject,
                                  RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype) do begin

          if ReferenceTypeOf(RdfSubject) = rtLocal then begin
            if NodeIDMap.IndexOfName(RdfSubject) = -1 then begin
              NewRdfSubject := Target.GetNewNodeID;
              NodeIDMap.Add(RdfSubject + '=' + newRdfSubject);
              RdfSubject := NewRdfSubject;
            end else
              RdfSubject := NodeIDMap.Values[RdfSubject];
          end;

          if ReferenceTypeOf(RdfObject) = rtLocal then begin
            if NodeIDMap.IndexOfName(RdfObject) = -1 then begin
              NewRdfObject := Target.GetNewNodeID;
              NodeIDMap.Add(RdfObject + '=' + newRdfObject);
              RdfObject := NewRdfObject;
            end else
              RdfObject := NodeIDMap.Values[RdfObject];
          end;

          Target.Add(RdfSubject, RdfPredicate, RdfObject,
                     RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype);

        end;
      finally
        Target.EnableControls;
      end;
    finally
      NodeIDMap.Free;
      TripleAnalyzer.Free;
      TripleTokenizer.Free;
      LineTokenizer.Free;
      InputSource.Free;
    end;
    Result := True;
  except
    Result := False;
  end;
end;

function TRdfNTripleToRdfGraphParser.ParseString(const Expression: string;
  var Handle: Integer): Boolean;
var
  StrStream: TStringStream;
begin
  StrStream:= TStringStream.Create(Expression);
  try
    Result:= ParseStream(StrStream, Handle);
  finally
    StrStream.free;
  end; 
end;

procedure TRdfNTripleToRdfGraphParser.SetTarget(
  Value: TRdfFilteredGraph);
begin
  FTarget := Value;
  if Assigned(Value) then Value.FreeNotification(Self);
end;

end.
