unit RdfSerializer_1_0;

// RDF Serializer 1.0.2
// Delphi 2010 Implementation
//
//
// LICENSE
//
// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// "http://www.mozilla.org/MPL/"
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
// the specific language governing rights and limitations under the License.
//
// The Original Code is "RdfSerializer_1_0.pas".
//
// The Initial Developer of the Original Code is Dieter Khler (Heidelberg,
// Germany, "http://www.philo.de/"). Portions created by the Initial Developer
// are Copyright (C) 2004-2010 Dieter Khler. All Rights Reserved.
//
// Alternatively, the contents of this file may be used under the terms of the
// GNU General Public License Version 2 or later (the "GPL"), in which case the
// provisions of the GPL are applicable instead of those above. If you wish to
// allow use of your version of this file only under the terms of the GPL, and
// not to allow others to use your version of this file under the terms of the
// MPL, indicate your decision by deleting the provisions above and replace them
// with the notice and other provisions required by the GPL. If you do not delete
// the provisions above, a recipient may use your version of this file under the
// terms of any one of the MPL or the GPL.


// HISTORY
//
// Only major modification are listed here.
//
// 2010-05-11 Release of version 1.0.2
// 2009-02-25 Release of version 1.0.1
// 2004-12-11 Release of version 1.0.0

interface

uses
  RDF_1_1, Classes;

type
  TRdfNTripleToRdfGraphSerializer = class(TComponent)
  private
    FSource: TRdfFilteredGraph;
    procedure SetSource(Value: TRdfFilteredGraph);
  protected
    procedure Notification(AComponent: TComponent;
                           Operation: TOperation); override;
  public
    function SerializeFile(const Filename: string): Boolean; virtual;
    function SerializeStream(const Stream: TStream): Boolean; virtual;
    function SerializeString(var S: string): Boolean; virtual;
  published
    property Source: TRdfFilteredGraph read FSource write SetSource;
  end;

procedure Register;

implementation

uses
  RdfConst, dkParserUtilsWin32, dkCodecUtilsWin32, SysUtils;

procedure Register;
begin
  RegisterComponents('Open RDF',[ TRdfNTripleToRdfGraphSerializer ]);
end;

resourcestring
  SCPOutOfRange = 'Code point (Unicode scalar value) out of range.';

const
  STREAM_BUF_SIZE = 4096;  // Buffer size for serialization stream
  UCS4_STRING_TERMINATOR = $9C;

type
  TRdfOutputSource = class(TUtilsCustomOutputStream)
  private
    procedure WriteUCS4Char(const C: UCS4Char);
  public
    procedure WriteObjectLiteral(const S: string);
    procedure WriteString(const AString: string);
    property BufSize;
  end;

{ TRdfOutputSource }

procedure TRdfOutputSource.WriteObjectLiteral(const S: string);
var
  StringStream: TStringStream;
  UCS4Reader: TUtilsUCS4Reader;
begin
  StringStream := TStringStream.Create(S);
  try
    UCS4Reader := TUtilsUCS4Reader.Create(StringStream, STREAM_BUF_SIZE,
                    'UTF-8', 0, 0, 0, 0, 1);
    try
      UCS4Reader.Next;
      while UCS4Reader.CurrentCharInfo.CodePoint <> UCS4_STRING_TERMINATOR do begin
        WriteUCS4Char(UCS4Reader.CurrentCharInfo.CodePoint);
        UCS4Reader.Next;
      end;
    finally
      UCS4Reader.Free;
    end;
  finally
    StringStream.Free;
  end;
end;

procedure TRdfOutputSource.WriteString(const AString: string);
begin
  Write(PChar(AString)^, Length(AString));
end;

procedure TRdfOutputSource.WriteUCS4Char(const C: UCS4Char);
var
  S: string;
begin
  case C of
    $9: S := '\t';
    $A: S := '\n';
    $D: S := '\r';
    $22: S := '\';
    $5C: S := '\\';
    $20..$21, $23..$5B, $5D..$7E:
      S := Char(C);
    $0..$8, $B, $C, $F..$1F, $7F..$FFFF:
      S := '\u' + IntToHex(C, 4);
  else
    if C > $10FFFF then
      raise EConvertError.Create(SCPOutOfRange);
      S := '\U' + IntToHex(C, 8);
  end;
  WriteString(S);
end;

{ TRdfNTripleToRdfGraphSerializer }

procedure TRdfNTripleToRdfGraphSerializer.Notification(
  AComponent: TComponent; Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent = FSource)
    then FSource := nil;
end;

function TRdfNTripleToRdfGraphSerializer.SerializeFile(
  const Filename: string): Boolean;
var
  FileStream: TFileStream;
begin
  FileStream:= TFileStream.create(Filename, fmCreate);
  try
    Result := SerializeStream(FileStream);
  finally
    FileStream.Free;
  end;
end;

function TRdfNTripleToRdfGraphSerializer.SerializeStream(
  const Stream: TStream): Boolean;
var
  I: Integer;
  Writer: TRdfOutputSource;
  DataType: string;
  Lang: string;
  Obj: string;
begin
  Writer := TRdfOutputSource.Create(Stream, STREAM_BUF_SIZE);
  try
    for I := 0 to Pred(Source.Count) do begin
      Writer.WriteString(Source.GetStatementAttribute(I, saSubject));
      Writer.WriteString(' ');
      Writer.WriteString(Source.GetStatementAttribute(I, saPredicate));
      Writer.WriteString(' ');
      Obj := Source.GetStatementAttribute(I, saObject);
      if Obj = '' then begin

        Writer.WriteString('"');
        Writer.WriteObjectLiteral(Source.GetStatementAttribute(I, saObjectLexical));
        Writer.WriteString('"');

        Lang := Source.GetStatementAttribute(I, saObjectLanguage);
        if Lang <> '' then
          Writer.WriteString('@' + Lang);

        DataType := Source.GetStatementAttribute(I, saObjectDatatype);
        if DataType <> '' then
          Writer.WriteString('^^' + DataType);

      end else
        Writer.WriteString(Obj);
      Writer.WriteString(' .'#$D#$A);
    end;
    Result := True;
  finally
    Writer.Free;
  end;
end;

function TRdfNTripleToRdfGraphSerializer.SerializeString(
  var S: string): Boolean;
var
  StringStream: TStringStream;
begin
  StringStream:= TStringStream.create('');
  try
    Result := SerializeStream(StringStream);
    S := StringStream.DataString;
  finally
    StringStream.Free;
  end;
end;

procedure TRdfNTripleToRdfGraphSerializer.SetSource(
  Value: TRdfFilteredGraph);
begin
  FSource := Value;
  if Assigned(Value) then Value.FreeNotification(Self);
end;

end.
