unit Rdf_1_1;

// RDF 1.1.7
// Delphi 2010 Implementation
//
//
// LICENSE
//
// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// "http://www.mozilla.org/MPL/"
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
// the specific language governing rights and limitations under the License.
//
// The Original Code is "Rdf_1_1.pas".
//
// The Initial Developer of the Original Code is Dieter Khler (Heidelberg,
// Germany, "http://www.philo.de/"). Portions created by the Initial Developer
// are Copyright (C) 2003-2010 Dieter Khler. All Rights Reserved.
//
// Alternatively, the contents of this file may be used under the terms of the
// GNU General Public License Version 2 or later (the "GPL"), in which case the
// provisions of the GPL are applicable instead of those above. If you wish to
// allow use of your version of this file only under the terms of the GPL, and
// not to allow others to use your version of this file under the terms of the
// MPL, indicate your decision by deleting the provisions above and replace them
// with the notice and other provisions required by the GPL. If you do not delete
// the provisions above, a recipient may use your version of this file under the
// terms of any one of the MPL or the GPL.


// HISTORY
//
// Only major modifications are listed here.
//
// 2010-05-11 Release of version 1.1.7
// 2009-02-25 Release of version 1.1.6
// 2004-06-01 Release of version 1.1.5
//            TRdfStatement.Handle property removed.
//            Other methods adapted accordingly.
// 2004-02-26 Release of version 1.1.4
// 2003-10-20 Release of version 1.1.3
//            TRdfCustomListGraph: Automatic filters added.
//            Access to graph abbreviations modified.
//            TRdfResourceList added.
// 2003-05-03 Release of version 1.1.1
//            TRdfDetailGraph added.
//            TRdfMemoryListGraph.SortType property added.
// 2003-04-01 Release of version 1.1.0
// 2003-03-26 Release of version 1.0.0

interface

uses
  RdfConst, AutoListUtils, dkUriUtils, dkParserUtilsWin32,
  StrUtils, SysUtils, Classes;

type
  UTF8String = string;

  ERdfException = class(Exception);

{ Forward declarations }

  TRdfStatement = class;
  TRdfGraph = class;
  TRdfGraphIterator = class;
  TRdfIteratorAdapter = class;

  TRdfReferenceType = (rtNone, rtGlobal, rtLocal);

  TRdfStmtAttr = (saSubject, saPredicate, saObject,
                  saObjectLexical, saObjectLanguage, saObjectDatatype);

  TRdfStmtAttrs = set of TRdfStmtAttr;

{ Event prototypes }

  TRdfFilterStatementEvent = procedure(Sender: TObject; RdfSubject,
    RdfPredicate, RdfObject, RdfObjectLexical, RdfObjectLanguage,
    RdfObjectDatatype: UTF8String; var Accept: Boolean) of object;
  TRdfFilterStringEvent = procedure(Sender: TObject; var S: string;
    var Accept: Boolean) of object;
  TRdfStringEvent = procedure(Sender: TObject; var S: string) of object;

{ TRdfStatement }

  TRdfStatement = class(TPersistent)
  private
    FObject_Datatype: UTF8String;
    FObject_Language: UTF8String;
    FObject_Lexical: UTF8String;
    FObject_Reference: UTF8String;
    FOwner: TObject;
    FPredicate_Reference: UTF8String;
    FSubject_Reference: UTF8String;
  public
    constructor Create(AOwner: TObject;
                       ASubject_Reference,
                       APredicate_Reference,
                       AObject_Reference,
                       AObject_Lexical,
                       AObject_Language,
                       AObject_Datatype: UTF8String);
    procedure Assign(Source: TPersistent); override;
    function GetAttribute(Kind: TRdfStmtAttr): UTF8String;

    property Object_Datatype: UTF8String read FObject_Datatype;
    property Object_Language: UTF8String read FObject_Language;
    property Object_Lexical: UTF8String read FObject_Lexical;
    property Object_Reference: UTF8String read FObject_Reference;
    property Owner: TObject read FOwner;
    property Predicate_Reference: UTF8String read FPredicate_Reference;
    property Subject_Reference: UTF8String read FSubject_Reference;
  end;

  TRdfCustomFilter = class
  protected
    function GetQualifier: string; virtual; abstract;
    procedure SetQualifier(Value: string); virtual; abstract;
    function UriMatch(const Expression,
                            URI: string): Boolean;
  public
    property Qualifier: string read GetQualifier write SetQualifier;
  end;

  TRdfUriFilter = class(TRdfCustomFilter)
  private
    FQualifier: string;
  protected
    function GetQualifier: string; override;
    procedure SetQualifier(Value: string); override;
  public
    constructor Create;
    function Match(AUri: UTF8String): Boolean; virtual;
  end;

  TRdfObjectFilter = class(TRdfCustomFilter)
  private
    FDatatypeQualifier: string;
    FLanguageQualifier: string;
    FLexicalQualifier: string;
    FReferenceQualifier: string;
  protected
    function GetQualifier: string; override;
    function MatchDataType(const Value: UTF8String): Boolean; virtual;
    function MatchLanguage(const Value: UTF8String): Boolean; virtual;
    function MatchLexical(const Value: UTF8String): Boolean; virtual;
    function MatchReference(const Value: UTF8String): Boolean; virtual;
    procedure ParseQualifier(const Value: string; out LexicalQualifier,
      LanguageQualifier, DatatypeQualifier: string); virtual;
    procedure SetQualifier(Value: string); override;
  public
    constructor Create;
    function Match(AObjectUri,
                   AObjectLexical,
                   AObjectLanguage,
                   AObjectDatatype: UTF8String): Boolean; virtual;
  end;

  TRdfStatementFilter = class(TPersistent)
  private
    FEnabled: Boolean;
    FObjectFilter: TRdfObjectFilter;
    FPredicateFilter: TRdfUriFilter;
    FRejectAll: Boolean;
    FSubjectFilter: TRdfUriFilter;
    function GetObjectQualifier: string;
    function GetPredicateQualifier: string;
    function GetSubjectQualifier: string;
    procedure SetObjectQualifier(Value: string);
    procedure SetPredicateQualifier(Value: string);
    procedure SetSubjectQualifier(Value: string);
  protected
    procedure SetEnabled(const Value: Boolean); virtual;
    procedure SetRejectAll(const Value: Boolean); virtual;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    function Match(ASubject,
                   APredicate,
                   AObject,
                   AObjectLexical,
                   AObjectLanguage,
                   AObjectDatatype: UTF8String): Boolean; virtual;
  published
    property Enabled: Boolean read FEnabled write SetEnabled default True;
    property ObjectQualifier: string read GetObjectQualifier write SetObjectQualifier;
    property PredicateQualifier: string read GetPredicateQualifier write SetPredicateQualifier;
    property RejectAll: Boolean read FRejectAll write SetRejectAll default False;
    property SubjectQualifier: string read GetSubjectQualifier write SetSubjectQualifier;
  end;

{ TRdfCustomGraphList }

  TRdfCustomGraphList = class(TUtilsCustomAliasedStrings)
  private
    FAdapter: TRdfIteratorAdapter;
    FGraphIterator: TRdfGraphIterator;
    function GetRdfGraph: TRdfGraph;
    procedure SetRdfGraph(Value: TRdfGraph);
  protected
    // Event Handlers:
    procedure ActiveChangedEventHandler(Sender: TObject); virtual;
    procedure ListModifiedEventHandler(Sender: TObject); virtual;

    property GraphIterator: TRdfGraphIterator read FGraphIterator;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    // Access:
    function HasStr(const S: string): Boolean; override;
    function IndexOf(const S: string; I: Integer): Integer; override;
    // Access via literal value:
    function HasLiteralStr(const S: string): Boolean; override;
    function IndexOfLiteral(const S: string; I: Integer): Integer; override;
    // View Control:
    procedure DisableControls; virtual;
    procedure EnableControls; virtual;

    property Count;
  published
    property Aliases;
    property AliasOpt;
    property RdfGraph: TRdfGraph read GetRdfGraph write SetRdfGraph;
  end;

{ TRdfGraphList }

  TRdfGraphList = class(TRdfCustomGraphList)
  private
    FRdfStatementAttribute: TRdfStmtAttr;
    FOnGetUserText: TRdfStringEvent;
    procedure SetRdfStatementAttribute(const Value: TRdfStmtAttr);
  protected
    procedure DoOnGetUserText(var S: string); virtual;
    procedure UpdateGraphList; virtual;
    // Event Handlers:
    procedure ActiveChangedEventHandler(Sender: TObject); override;
    procedure ListModifiedEventHandler(Sender: TObject); override;
  published
    property StatementAttribute: TRdfStmtAttr read FRdfStatementAttribute write SetRdfStatementAttribute;

    property OnGetUserText: TRdfStringEvent read FOnGetUserText write FOnGetUserText;
  end;

{ TRdfResourceList }

  TRdfResourceListModus = set of (rlmSubjectResources, rlmPredicateResources,
                                  rlmObjectResources, rlmDatatypeResources);

  TRdfResourceList = class(TRdfCustomGraphList)
  private
    FModus: TRdfResourceListModus;
    FOnGetUserText: TRdfFilterStringEvent;
  protected
    procedure DoOnGetUserText(var S: string;
                              var Accept: Boolean); virtual;
    procedure SetModus(Value: TRdfResourceListModus); virtual;
    procedure UpdateResourceList; virtual;
    // Event Handlers:
    procedure ActiveChangedEventHandler(Sender: TObject); override;
    procedure ListModifiedEventHandler(Sender: TObject); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Modus: TRdfResourceListModus read FModus write SetModus;

    property OnGetUserText: TRdfFilterStringEvent read FOnGetUserText write FOnGetUserText;
  end;

{ TRdfIteratorAdapter }

  TRdfIteratorAdapter = class(TUtilsIteratorAdapter)
  protected
    function GetIterator: TRdfGraphIterator; reintroduce;
    procedure SetIterator(AIterator: TRdfGraphIterator); reintroduce;
  public
    property Iterator: TRdfGraphIterator read GetIterator write SetIterator;
  end;

{ TRdfGraph }

  TRdfSortType = (stNone, stSPO, stSOP, stPSO, stPOS, stOSP, stOPS);

  TRdfGraph = class(TUtilsCustomObjectIteratorList)
  private
    FLastNodeID: Integer;
    FSortType: TRdfSortType;
    function GetObject(Index: Integer): UTF8String;
    function GetObjectDatatype(Index: Integer): UTF8String;
    function GetObjectLanguage(Index: Integer): UTF8String;
    function GetObjectLexical(Index: Integer): UTF8String;
    function GetPredicate(Index: Integer): UTF8String;
    function GetStatement(Index: Integer): TRdfStatement;
    function GetSubject(Index: Integer): UTF8String;
    procedure PutStatement(Index: Integer; const Value: TRdfStatement);
  protected
    function DoCompare(const Object1, Object2: TObject): Integer; override;
    function EscapeStr(S: string): string; virtual;
    function Find(const Stmt: TRdfStatement;
                    var Index: Integer): Boolean; reintroduce; virtual;
    function GetNewNodeID: UTF8String; virtual;
    procedure SetSortType(Value: TRdfSortType); virtual;
    function UnescapeStr(S: string): string; virtual;

    // List mutation:
    function Add(RdfSubject,
                 RdfPredicate,
                 RdfObject,
                 RdfObjectLexical,
                 RdfObjectLanguage,
                 RdfObjectDatatype: UTF8String): Integer; reintroduce; virtual;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    function Edit(Index: Integer;
                  RdfSubject,
                  RdfPredicate,
                  RdfObject,
                  RdfObjectLexical,
                  RdfObjectLanguage,
                  RdfObjectDatatype: UTF8String): Integer; virtual;
    function Insert(Index: Integer;
                    RdfSubject,
                    RdfPredicate,
                    RdfObject,
                    RdfObjectLexical,
                    RdfObjectLanguage,
                    RdfObjectDatatype: UTF8String): Boolean; reintroduce; virtual;

    property SortType: TRdfSortType read FSortType write SetSortType default stNone;
    property Statements[Index: Integer]: TRdfStatement read GetStatement write PutStatement;

    property Subjects[Index: Integer]: UTF8String read GetSubject;
    property Predicates[Index: Integer]: UTF8String read GetPredicate;
    property Objects[Index: Integer]: UTF8String read GetObject;
    property ObjectLexicals[Index: Integer]: UTF8String read GetObjectLexical;
    property ObjectLanguages[Index: Integer]: UTF8String read GetObjectLanguage;
    property ObjectDatatypes[Index: Integer]: UTF8String read GetObjectDatatype;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    // List access:
    function GetStatementAttribute(Index: Integer;
                                   Kind: TRdfStmtAttr): string; virtual;
    // View Control:
    procedure DisableControls; virtual;
    procedure EnableControls; virtual;
  end;

{ TRdfCustomFilteredGraph }

  TRdfCustomFilteredGraph = class(TRdfGraph)
  private
    FFilter: TRdfStatementFilter;
    FOnFilterStatement: TRdfFilterStatementEvent;
    FAutoFilter: Boolean;
  protected
    procedure DoOnFilterStatement(    RdfSubject,
                                      RdfPredicate,
                                      RdfObject,
                                      RdfObjectLexical,
                                      RdfObjectLanguage,
                                      RdfObjectDatatype: UTF8String;
                                  out Accept: Boolean); virtual;
    function MatchFilter(RdfSubject,
                         RdfPredicate,
                         RdfObject,
                         RdfObjectLexical,
                         RdfObjectLanguage,
                         RdfObjectDatatype: UTF8String): Boolean; virtual;
    procedure SetAutoFilter(const Value: Boolean); virtual;
    procedure SetFilter(const Value: TRdfStatementFilter); virtual;

    // List mutation:
    function Add(RdfSubject,
                 RdfPredicate,
                 RdfObject,
                 RdfObjectLexical,
                 RdfObjectLanguage,
                 RdfObjectDatatype: UTF8String): Integer; override;
    function Edit(Index: Integer;
                  RdfSubject,
                  RdfPredicate,
                  RdfObject,
                  RdfObjectLexical,
                  RdfObjectLanguage,
                  RdfObjectDatatype: UTF8String): Integer; override;
    function Insert(Index: Integer;
                    RdfSubject,
                    RdfPredicate,
                    RdfObject,
                    RdfObjectLexical,
                    RdfObjectLanguage,
                    RdfObjectDatatype: UTF8String): Boolean; override;

    property AutoFilter: Boolean read FAutoFilter write SetAutoFilter default false;
    property Filter: TRdfStatementFilter read FFilter write SetFilter;

    property OnFilterStatement: TRdfFilterStatementEvent read FOnFilterStatement write FOnFilterStatement;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

{ TRdfFilteredGraph }

  TRdfFilteredGraph = class(TRdfCustomFilteredGraph)
  public
    // List mutation:
    function Add(RdfSubject,
                 RdfPredicate,
                 RdfObject,
                 RdfObjectLexical,
                 RdfObjectLanguage,
                 RdfObjectDatatype: UTF8String): Integer; override;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    function Edit(Index: Integer;
                  RdfSubject,
                  RdfPredicate,
                  RdfObject,
                  RdfObjectLexical,
                  RdfObjectLanguage,
                  RdfObjectDatatype: UTF8String): Integer; override;
    function GetNewNodeID: UTF8String; override;
    function Insert(Index: Integer;
                    RdfSubject,
                    RdfPredicate,
                    RdfObject,
                    RdfObjectLexical,
                    RdfObjectLanguage,
                    RdfObjectDatatype: UTF8String): Boolean; override;

    property CanModify;
    property Count;
  published
    property Aliases;
    property AutoFilter;
    property Duplicates;
    property Filter;
    property Sorted;
    property SortType;

    property OnFilterStatement;
  end;

{ TRdfCustomDetailGraph }

  TRdfCustomDetailGraph = class(TRdfCustomFilteredGraph)
  private
    FAdapter: TRdfIteratorAdapter;
    FGraphIterator: TRdfGraphIterator;
    FOnAfterUpdateView: TNotifyEvent;
    FOnBeforeUpdateView: TNotifyEvent;
    function GetRdfGraph: TRdfGraph;
    procedure SetRdfGraph(Value: TRdfGraph);
  protected
    // Event Handlers:
    procedure ActiveChangedEventHandler(Sender: TObject); virtual;
    procedure ListModifiedEventHandler(Sender: TObject); virtual;

    procedure DoOnAfterUpdateView; virtual;
    procedure DoOnBeforeUpdateView; virtual;
    procedure UpdateView; virtual;

    property GraphIterator: TRdfGraphIterator read FGraphIterator;
    property RdfGraph: TRdfGraph read GetRdfGraph write SetRdfGraph;

    property OnAfterUpdateView: TNotifyEvent read FOnAfterUpdateView write FOnAfterUpdateView;
    property OnBeforeUpdateView: TNotifyEvent read FOnBeforeUpdateView write FOnBeforeUpdateView;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

{ TRdfDetailGraph }

  TRdfDetailGraph = class(TRdfCustomDetailGraph)
  public
    procedure ForceUpdate; virtual;
    property Count;
  published
    property AutoFilter;
    property Filter;
    property RdfGraph;

    property OnAfterUpdateView;
    property OnBeforeUpdateView;
    property OnFilterStatement;
  end;

{ TRdfCustomRestricterGraph }

  TRdfCustomRestricterGraph = class(TRdfCustomDetailGraph)
  private
    FRestricterFilter: TRdfStatementFilter;
    FSubjectRestricterAdapter: TUtilsStringIteratorAdapter;
    FPredicateRestricterAdapter: TUtilsStringIteratorAdapter;
    FObjectRestricterAdapter: TUtilsStringIteratorAdapter;
    function GetSubjectRestricter: TUtilsStringIterator;
    function GetPredicateRestricter: TUtilsStringIterator;
    function GetObjectRestricter: TUtilsStringIterator;
    procedure SetSubjectRestricter(Value: TUtilsStringIterator);
    procedure SetPredicateRestricter(Value: TUtilsStringIterator);
    procedure SetObjectRestricter(Value: TUtilsStringIterator);
  protected
    procedure ItemChangedEventHandler(Sender: TObject); virtual;
    procedure Notification(AComponent: TComponent;
                           Operation: TOperation); override;
    function MatchFilter(RdfSubject,
                         RdfPredicate,
                         RdfObject,
                         RdfObjectLexical,
                         RdfObjectLanguage,
                         RdfObjectDatatype: UTF8String): Boolean; override;
    procedure UpdateView; override;
    property RestricterFilter: TRdfStatementFilter read FRestricterFilter;

    property SubjectRestricter: TUtilsStringIterator read GetSubjectRestricter write SetSubjectRestricter;
    property PredicateRestricter: TUtilsStringIterator read GetPredicateRestricter write SetPredicateRestricter;
    property ObjectRestricter: TUtilsStringIterator read GetObjectRestricter write SetObjectRestricter;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

{ TRdfRestricterGraph }

  TRdfRestricterGraph = class(TRdfCustomRestricterGraph)
  public
    procedure ForceUpdate; virtual;
    property Count;
  published
    property AutoFilter;
    property Filter;
    property ObjectRestricter;
    property PredicateRestricter;
    property RdfGraph;
    property SubjectRestricter;

    property OnAfterUpdateView;
    property OnBeforeUpdateView;
    property OnFilterStatement;
  end;

{ TRdfGraphIterator }

  TRdfGraphIterator = class(TUtilsCustomIterator)
  private
    FRdfObject: UTF8String;
    FRdfObjectDatatype: UTF8String;
    FRdfObjectLanguage: UTF8String;
    FRdfObjectLexical: UTF8String;
    FRdfPredicate: UTF8String;
    FRdfSubject: UTF8String;
  protected
    function GetCanModify: Boolean; virtual;
    function GetRdfObject: UTF8String; virtual;
    function GetRdfObjectDatatype: UTF8String; virtual;
    function GetRdfObjectLanguage: UTF8String; virtual;
    function GetRdfObjectLexical: UTF8String; virtual;
    function GetRdfPredicate: UTF8String; virtual;
    function GetRdfSubject: UTF8String; virtual;

    function GetRdfGraph: TRdfGraph;
    procedure SetRdfGraph(const Value: TRdfGraph);

    procedure UpdateIterator(const NewPos: Integer;
                             const ItemChanged,
                                   ListModified: Boolean); override;
  public
    // Iterator Value Inquiring:
    property RdfSubject: UTF8String read GetRdfSubject;
    property RdfPredicate: UTF8String read GetRdfPredicate;
    property RdfObject: UTF8String read GetRdfObject;
    property RdfObjectLexical: UTF8String read GetRdfObjectLexical;
    property RdfObjectLanguage: UTF8String read GetRdfObjectLanguage;
    property RdfObjectDatatype: UTF8String read GetRdfObjectDatatype;

    property CanModify: Boolean read GetCanModify;
  published
    property RdfGraph: TRdfGraph read GetRdfGraph write SetRdfGraph;
  end;

{ TRdfAliases }

  TRdfAliases = class(TUtilsCustomIteratorAliases)
  private
    FNameAdapter: TRdfIteratorAdapter;
    FNameIterator: TRdfGraphIterator;
    FAliasAdapter: TRdfIteratorAdapter;
    FAliasIterator: TRdfGraphIterator;
    FOnAfterUpdate: TNotifyEvent;
    FOnBeforeUpdate: TNotifyEvent;
    FPredicateSourceAdapter: TUtilsStringIteratorAdapter;
    FPredicateURI: UTF8String;
    function GetNameSource: TRdfGraph;
    function GetAliasSource: TRdfGraph;
    function GetPredicateSource: TUtilsStringIterator;
    procedure SetNameSource(const Value: TRdfGraph);
    procedure SetAliasSource(const Value: TRdfGraph);
    procedure SetPredicateSource(const Value: TUtilsStringIterator);
    procedure SetPredicateURI(const Value: UTF8String);

    function IsPredicateURI(Value: UTF8String): Boolean;
  protected
    // Event Handlers:
    procedure GeneralEventHandler(Sender: TObject); virtual;

    procedure DoOnAfterUpdate; virtual;
    procedure DoOnBeforeUpdate; virtual;
    procedure Notification(AComponent: TComponent;
                           Operation: TOperation); override;
    procedure UpdateAliases; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property NameSource: TRdfGraph read GetNameSource write SetNameSource;
    property PredicateSource: TUtilsStringIterator read GetPredicateSource write SetPredicateSource;
    property PredicateURI: UTF8String read FPredicateURI write SetPredicateURI;
    property AliasSource: TRdfGraph read GetAliasSource write SetAliasSource;

    property OnAfterUpdate: TNotifyEvent read FOnAfterUpdate write FOnAfterUpdate;
    property OnBeforeUpdate: TNotifyEvent read FOnBeforeUpdate write FOnBeforeUpdate;
  end;

{ Global Methods }

function ReferenceTypeOf(S: string): TRdfReferenceType;

procedure Register;

resourcestring
  SInvalidFilterQualifier = 'Invalid statement filter qualifier';

implementation

uses
  dkCodecUtilsWin32;

{ Global Methods }

procedure Register;
begin
  RegisterComponents('Open RDF',[ TRdfAliases,
                             TRdfFilteredGraph,
                             TRdfDetailGraph,
                             TRdfRestricterGraph,
                             TRdfGraphIterator,
                             TRdfGraphList,
                             TRdfResourceList ]);
end;

function ReferenceTypeOf(S: string): TRdfReferenceType;
// This function does not test the well-formedness of S.
// xxx Make well-formedness test? xxx
begin
  Result := rtNone;
  if S <> '' then
    if S[1] = '<' then Result := rtGlobal
      else if S[1] = '_' then Result := rtLocal;
end;

{ Local Methods }

{ URI processing }

function IsUriStr(S: string): Boolean;
// Tests for '#' and uric production of RFC 2396, App. A.
var
  I, L: Integer;
begin
  L := Length(S);
  if L = 0
    then begin Result := False; Exit; end
    else Result := True;
  I := 0;
  while I < L do begin
    Inc(I);
    if S[I] = '%' then begin
      if I = L then begin Result := False; Exit; end;
      Inc(I);
      if not IsUriHexChar(S[I]) then begin Result := False; Exit; end;
      if I = L then begin Result := False; Exit; end;
      Inc(I);
      if not IsUriHexChar(S[I]) then begin Result := False; Exit; end;
    end else if not ( IsUriReservedChar(S[I]) or IsUriUnreservedChar(S[I]) or (S[I] = '#'))
      then begin Result := False; Exit; end;
  end;
end;

{ N-Triple syntax testing }

function IsNTripleNodeID(const S: string): Boolean;
var
  I: Integer;
begin
  if Length(S) < 3 then begin
    Result := False;
    Exit;
  end;
  if (S[1] <> '_') or (S[2] <> ':') or not IsUriAlphaChar(S[3]) then begin
    Result := False;
    Exit;
  end;
  Result := True;
  for I := 4 to Length(S) do
    if not (IsUriAlphaChar(S[I]) or IsUriDigitChar(S[I])) then begin
      Result := False;
      Exit;
    end;
end;

function IsNTripleUriRef(const S: string): Boolean;
begin
  if Length(S) < 3 then begin
    Result := False;
    Exit;
  end;
  if (S[1] <> '<') or (S[Length(S)] <> '>') then begin
    Result := False;
    Exit;
  end;
  Result := IsUriStr(Copy(S, 2, Length(S) - 2));
end;

{ TRdfStatement }

constructor TRdfStatement.Create(AOwner: TObject;
                                 ASubject_Reference,
                                 APredicate_Reference,
                                 AObject_Reference,
                                 AObject_Lexical,
                                 AObject_Language,
                                 AObject_Datatype: UTF8String);
begin
  inherited Create;
  FObject_Datatype := AObject_Datatype;
  FObject_Language := AObject_Language;
  FObject_Lexical := AObject_Lexical;
  FObject_Reference := AObject_Reference;
  FOwner := AOwner;
  FPredicate_Reference := APredicate_Reference;
  FSubject_Reference := ASubject_Reference;
end;

procedure TRdfStatement.Assign(Source: TPersistent);
begin
  if Source is TRdfStatement then begin
    FObject_Datatype     := TRdfStatement(Source).FObject_Datatype;
    FObject_Language     := TRdfStatement(Source).FObject_Language;
    FObject_Lexical      := TRdfStatement(Source).FObject_Lexical;
    FObject_Reference    := TRdfStatement(Source).FObject_Reference;
    FPredicate_Reference := TRdfStatement(Source).FPredicate_Reference;
    FSubject_Reference   := TRdfStatement(Source).FSubject_Reference;
    exit;
  end;
  inherited Assign(Source);
end;

function TRdfStatement.GetAttribute(Kind: TRdfStmtAttr): UTF8String;
begin
  case Kind of
    saSubject:         Result := FSubject_Reference;
    saPredicate:       Result := FPredicate_Reference;
    saObject:          Result := FObject_Reference;
    saObjectLexical:   Result := FObject_Lexical;
    saObjectLanguage:  Result := FObject_Language;
    saObjectDatatype:  Result := FObject_Datatype;
  end;
end;

{ TRdfCustomFilter }

function TRdfCustomFilter.UriMatch(const Expression, URI: string): Boolean;
begin
  if (Expression = '*') or (Expression = '<*>') then
    begin Result := True; Exit; end;
  if Expression = URI then
    begin Result := True; Exit; end;
  if Length(Expression) > 1 then begin
    if Expression[pred(Length(Expression))] = '*'
      then Result := Copy(Expression, 1, Length(Expression) - 2) =
                     Copy(URI, 1, Length(Expression) - 2)
      else Result := False;
  end else Result := False;
end;

{ TRdfUriFilter }

constructor TRdfUriFilter.Create;
begin
  FQualifier := '*';
end;

function TRdfUriFilter.GetQualifier: string;
begin
  Result := FQualifier;
end;

function TRdfUriFilter.Match(AUri: UTF8String): Boolean;
begin
  Result := UriMatch(Qualifier, AUri);
end;

procedure TRdfUriFilter.SetQualifier(Value: string);
begin
  Value := Trim(Value);
  if not (IsNTripleUriRef(Value) or IsNTripleNodeID(Value)) then
    raise ERdfException.Create(SInvalidFilterQualifier);
  FQualifier := Value;
end;

{ TRdfObjectFilter }

constructor TRdfObjectFilter.Create;
begin
  FReferenceQualifier := '*';
  FLexicalQualifier   := '\*';
  FLanguageQualifier  := '*';
  FDatatypeQualifier  := '*';
end;

function TRdfObjectFilter.GetQualifier: string;
begin
  if (FReferenceQualifier = '*')  and (FLexicalQualifier   = '\*') and
     (FLanguageQualifier  = '*')  and (FDatatypeQualifier  = '*') then
    begin Result := '*'; Exit; end;
  if FReferenceQualifier <> '' then
    begin Result := FReferenceQualifier; Exit; end;
  Result := '"' + FLexicalQualifier + '" @' + FLanguageQualifier;
  if FDatatypeQualifier <> '' then
    Result := Result + ' <' + FDatatypeQualifier +'>';
end;

function TRdfObjectFilter.Match(AObjectUri, AObjectLexical, AObjectLanguage,
  AObjectDatatype: UTF8String): Boolean;
begin
  Result := MatchReference(AObjectUri) and
            MatchLexical(AObjectLexical) and
            MatchLanguage(AObjectLanguage) and
            MatchDataType(AObjectDatatype);
end;

function TRdfObjectFilter.MatchDataType(const Value: UTF8String): Boolean;
begin
  if FDatatypeQualifier = ''
    then Result := Value = ''
    else Result := UriMatch('<' + FDatatypeQualifier + '>', Value);
end;

function TRdfObjectFilter.MatchLanguage(const Value: UTF8String): Boolean;
begin
  Result := (FLanguageQualifier = '*') or
            (FLanguageQualifier = '<*>') or
            (FLanguageQualifier = Value);
end;

function TRdfObjectFilter.MatchLexical(const Value: UTF8String): Boolean;
begin
  Result := (FLexicalQualifier = '\*') or
            (FLexicalQualifier = Value);
end;

function TRdfObjectFilter.MatchReference(const Value: UTF8String): Boolean;
begin
  if FReferenceQualifier = ''
    then Result := Value = ''
    else Result := UriMatch(FReferenceQualifier, Value);
end;

procedure TRdfObjectFilter.ParseQualifier(const Value: string;
                                            out LexicalQualifier,
                                                LanguageQualifier,
                                                DatatypeQualifier: string);
const
  STRING_TERMINATOR = $9C;
  WHITESPACE: WideString = #$0009#$0020;
var
  InputSource: TUtilsUCS4Reader;
  Stream: TStringStream;
begin
  Stream := nil;
  InputSource := nil;
  LexicalQualifier := '';
  LanguageQualifier := '';
  DatatypeQualifier := '';
  try
    Stream:= TStringStream.Create(Value);
    InputSource:= TUtilsUCS4Reader.Create(Stream, 4096, 'US-ASCII', 0, 0, 0, 0, 1);
    with InputSource do begin
      Next;
      if CurrentCharInfo.CodePoint <> $22 then // " ?
        raise ERdfException.Create(SInvalidFilterQualifier);
      Next;
      while CurrentCharInfo.CodePoint <> $22 do begin
        if CurrentCharInfo.CodePoint = STRING_TERMINATOR then
          raise ERdfException.Create(SInvalidFilterQualifier);
        LexicalQualifier := LexicalQualifier + Char(CurrentCharInfo.CodePoint);
        Next;
      end;
      // ### Add wellformedness check for LexicalQualifier!
      SkipNext(WHITESPACE);
      if CurrentCharInfo.CodePoint <> $40 then // @ ?
        raise ERdfException.Create(SInvalidFilterQualifier);
      repeat
        Next;
        if CurrentCharInfo.CodePoint = STRING_TERMINATOR then
          raise ERdfException.Create(SInvalidFilterQualifier);
        LanguageQualifier := LanguageQualifier + Char(CurrentCharInfo.CodePoint);
      until (CurrentCharInfo.CodePoint = $9) or (CurrentCharInfo.CodePoint = $20);  // Whitespace ?
      // ### Add wellformedness check for LanguageQualifier!
      SkipNext(WHITESPACE);
      if CurrentCharInfo.CodePoint = STRING_TERMINATOR then exit;
      repeat
        DatatypeQualifier := DatatypeQualifier + Char(CurrentCharInfo.CodePoint);
        Next;
      until (CurrentCharInfo.CodePoint = $9) or (CurrentCharInfo.CodePoint = $20) or
            (CurrentCharInfo.CodePoint = STRING_TERMINATOR);  // Whitespace or string end?
      // ### Add wellformedness check for DatatypeQualifier!
      SkipNext(WHITESPACE);
      if CurrentCharInfo.CodePoint <> STRING_TERMINATOR then
        raise ERdfException.Create(SInvalidFilterQualifier);
    end;
  finally
    InputSource.Free;
    Stream.Free;
  end;
end;

procedure TRdfObjectFilter.SetQualifier(Value: string);
// ### Add wellformedness checks!
var
  LitQlfTemp, LangQlfTemp, DatTypeQlfTemp: string;
begin
  Value := Trim(Value);
  if Value = '*' then begin     // Anything?
    FReferenceQualifier := '*';
    FLexicalQualifier   := '\*';
    FLanguageQualifier  := '*';
    FDatatypeQualifier  := '*';
    Exit;
  end;
  if Length(Value) < 2 then
    raise ERdfException.Create(SInvalidFilterQualifier);
  if IsNTripleUriRef(Value) or IsNTripleNodeID(Value) then begin  // Object reference?
    FReferenceQualifier := Value;
    FLexicalQualifier   := '';
    FLanguageQualifier  := '';
    FDatatypeQualifier  := '';
  end else begin
    ParseQualifier(Value, LitQlfTemp, LangQlfTemp, DatTypeQlfTemp);
      // We use temporary variables, in case that an exception occurs.
    if not IsNTripleUriRef(DatTypeQlfTemp) then
      raise ERdfException.Create(SInvalidFilterQualifier);
    FDatatypeQualifier  := DatTypeQlfTemp;
    FReferenceQualifier := '';
    FLexicalQualifier   := LitQlfTemp;
    FLanguageQualifier  := LangQlfTemp;
  end;
end;

{ TRdfStatementFilter }

constructor TRdfStatementFilter.Create;
begin
  inherited;
  FEnabled := True;
  FObjectFilter := TRdfObjectFilter.Create;
  FPredicateFilter := TRdfUriFilter.Create;
  FSubjectFilter := TRdfUriFilter.Create;
end;

destructor TRdfStatementFilter.Destroy;
begin
  FObjectFilter.Free;
  FPredicateFilter.Free;
  FSubjectFilter.Free;
  inherited;
end;

procedure TRdfStatementFilter.Assign(Source: TPersistent);
begin
  if Source is TRdfStatementFilter then begin
    ObjectQualifier    := TRdfStatementFilter(Source).ObjectQualifier;
    PredicateQualifier := TRdfStatementFilter(Source).PredicateQualifier;
    SubjectQualifier   := TRdfStatementFilter(Source).SubjectQualifier;
    exit;
  end;
  inherited Assign(Source);
end;

function TRdfStatementFilter.GetObjectQualifier: string;
begin
  Result := FObjectFilter.Qualifier;
end;

function TRdfStatementFilter.GetPredicateQualifier: string;
begin
  Result := FPredicateFilter.Qualifier;
end;

function TRdfStatementFilter.GetSubjectQualifier: string;
begin
  Result := FSubjectFilter.Qualifier;
end;

function TRdfStatementFilter.Match(ASubject,
  APredicate, AObject, AObjectLexical, AObjectLanguage,
  AObjectDatatype: UTF8String): Boolean;
begin
  if Enabled then begin
    if RejectAll
      then Result := False
      else Result := FSubjectFilter.Match(ASubject) and
                     FPredicateFilter.Match(APredicate) and
                     FObjectFilter.Match(AObject, AObjectLexical, AObjectLanguage, AObjectDatatype)
  end else Result := True;
end;

procedure TRdfStatementFilter.SetEnabled(const Value: Boolean);
begin
  FEnabled := Value;
end;

procedure TRdfStatementFilter.SetObjectQualifier(Value: string);
begin
  FObjectFilter.Qualifier := Value;
end;

procedure TRdfStatementFilter.SetPredicateQualifier(Value: string);
begin
  FPredicateFilter.Qualifier := Value;
end;

procedure TRdfStatementFilter.SetRejectAll(const Value: Boolean);
begin
  FRejectAll := Value;
end;

procedure TRdfStatementFilter.SetSubjectQualifier(Value: string);
begin
  FSubjectFilter.Qualifier := Value;
end;

{ TRdfCustomGraphList }

procedure TRdfCustomGraphList.ActiveChangedEventHandler(Sender: TObject);
begin
  // By default do nothing.
end;

constructor TRdfCustomGraphList.Create(AOwner: TComponent);
begin
  inherited;
  FGraphIterator := TRdfGraphIterator.Create(self);
  FAdapter := TRdfIteratorAdapter.Create;
  with FAdapter do begin
    SetIterator(FGraphIterator); // Setting the iterator must go before setting the event handlers.
    OnActiveChanged := ActiveChangedEventHandler;
    OnListModified := ListModifiedEventHandler;
  end;
  ReadOnly := True;
end;

destructor TRdfCustomGraphList.Destroy;
begin
  SetRdfGraph(nil);
  FAdapter.Free;
  FAdapter := nil;
  FGraphIterator.Free;
  inherited;
end;

procedure TRdfCustomGraphList.DisableControls;
begin
  DoNotifyDisableControls;
end;

procedure TRdfCustomGraphList.EnableControls;
begin
  DoNotifyEnableControls;
end;

function TRdfCustomGraphList.GetRdfGraph: TRdfGraph;
begin
  Result := GraphIterator.RdfGraph;
end;

function TRdfCustomGraphList.HasLiteralStr(const S: string): Boolean;
begin
  Result := inherited HasLiteralStr(S);
end;

function TRdfCustomGraphList.HasStr(const S: string): Boolean;
begin
  Result := inherited HasStr(S);
end;

function TRdfCustomGraphList.IndexOf(const S: string; I: Integer): Integer;
begin
  Result := inherited IndexOf(S, I);
end;

function TRdfCustomGraphList.IndexOfLiteral(const S: string;
  I: Integer): Integer;
begin
  Result := inherited IndexOfLiteral(S, I);
end;

procedure TRdfCustomGraphList.ListModifiedEventHandler(Sender: TObject);
begin
  // By default do nothing.
end;

procedure TRdfCustomGraphList.SetRdfGraph(Value: TRdfGraph);
begin
  if not (csLoading in GraphIterator.ComponentState) then
    GraphIterator.RdfGraph := Value;
end;

{ TRdfGraphList }

procedure TRdfGraphList.ActiveChangedEventHandler(Sender: TObject);
begin
  UpdateGraphList;
end;

procedure TRdfGraphList.DoOnGetUserText(var S: string);
begin
  if Assigned(FOnGetUserText) then FOnGetUserText(Self, S);
end;

procedure TRdfGraphList.ListModifiedEventHandler(Sender: TObject);
begin
  UpdateGraphList;
end;

procedure TRdfGraphList.SetRdfStatementAttribute(
  const Value: TRdfStmtAttr);
begin
  if FRdfStatementAttribute <> Value then begin
    FRdfStatementAttribute := Value;
    UpdateGraphList;
  end;
end;

(* xxx
procedure TRdfGraphList.UpdateFocus;
begin
  DisableControls;
  try
    with FRdfLink do
      if Active
        then Go(RdfGraph.StatementIndex + 1)
        else Go(0);
  finally
    EnableControls;
  end;
end;
      *)

procedure TRdfGraphList.UpdateGraphList;
var
  S: string;
begin
  DisableControls;
  try
    ReadOnly := False;
    Clear;
    with GraphIterator do begin
      First;
      while not Offright do begin
        S := RdfGraph.GetStatementAttribute(Position, StatementAttribute);
        DoOnGetUserText(S);
        AddLiteral(S);
        Next;
      end;
    end;
  finally
    ReadOnly := True;
    EnableControls;
  end;
end;

{ TRdfResourceList }

procedure TRdfResourceList.ActiveChangedEventHandler(Sender: TObject);
begin
  UpdateResourceList;
end;

constructor TRdfResourceList.Create(AOwner: TComponent);
var
  Backup: Boolean;
begin
  inherited;
  Backup := ReadOnly;
  ReadOnly := False;  // Necessary to prevent exception when setting Sorted.
  try
    Sorted := True;
    Duplicates := dupIgnore;
  finally
    ReadOnly := Backup;
  end;
end;

procedure TRdfResourceList.DoOnGetUserText(var S: string;
  var Accept: Boolean);
begin
  if Assigned(FOnGetUserText) then FOnGetUserText(Self, S, Accept);
end;

procedure TRdfResourceList.ListModifiedEventHandler(Sender: TObject);
begin
  UpdateResourceList;
end;

procedure TRdfResourceList.SetModus(Value: TRdfResourceListModus);
begin
  if FModus <> Value then begin
    FModus := Value;
    UpdateResourceList;
  end;
end;

procedure TRdfResourceList.UpdateResourceList;
var
  Accept: Boolean;
  S: string;
begin
  DisableControls;
  try
    ReadOnly := False;
    Clear;
    with GraphIterator do begin
      First;
      while not Offright do begin
        if rlmSubjectResources in Modus then begin
          S := RdfGraph.GetStatementAttribute(Position, saSubject);
          Accept := True;
          DoOnGetUserText(S, Accept);
          if Accept then AddLiteral(S);
        end;
        if rlmPredicateResources in Modus then begin
          S := RdfGraph.GetStatementAttribute(Position, saPredicate);
          Accept := True;
          DoOnGetUserText(S, Accept);
          if Accept then AddLiteral(S);
        end;
        if (rlmObjectResources in Modus) or (rlmDatatypeResources in Modus) then begin
          S := RdfGraph.GetStatementAttribute(Position, saObject);
          if S = '' then begin
            if rlmDatatypeResources in Modus then begin
              S := RdfGraph.GetStatementAttribute(Position, saObjectDatatype);
              Accept := True;
              DoOnGetUserText(S, Accept);
              if Accept then AddLiteral(S);
            end;
          end else begin
            if rlmObjectResources in Modus then begin
              Accept := True;
              DoOnGetUserText(S, Accept);
              if Accept then AddLiteral(S);
            end;
          end;
        end;
        Next;
      end;
    end;
  finally
    ReadOnly := True;
    EnableControls;
  end;
end;

{ TRdfIteratorAdapter }

function TRdfIteratorAdapter.GetIterator: TRdfGraphIterator;
begin
  Result := inherited GetIterator as TRdfGraphIterator;
end;

procedure TRdfIteratorAdapter.SetIterator(AIterator: TRdfGraphIterator);
begin
  inherited SetIterator(AIterator);
end;

{ TRdfGraph }

function TRdfGraph.Add(RdfSubject, RdfPredicate, RdfObject, RdfObjectLexical,
  RdfObjectLanguage, RdfObjectDatatype: UTF8String): Integer;
var
  NewStmt: TRdfStatement;
begin
  NewStmt := TRdfStatement.Create(Self,
                                  RdfSubject,
                                  RdfPredicate,                                  RdfObject,
                                  RdfObjectLexical,
                                  RdfObjectLanguage,
                                  RdfObjectDatatype);
  try
    Result:= inherited Add(newStmt);
  except
    NewStmt.Free;
    raise;
  end;
end;

procedure TRdfGraph.Clear;
var
  I : Integer;
begin
  CheckCanModify;
  for I := Pred(FItems.Count) downto 0 do
    TRdfStatement(FItems[I]).Free;
  FItems.Clear;
  DoNotifyClear;
end;

constructor TRdfGraph.Create(AOwner: TComponent);
begin
  inherited;
  FLastNodeID := 0;
  FSortType := stNone;
end;

procedure TRdfGraph.Delete(Index: Integer);
begin
  CheckCanModify;
  TRdfStatement(FItems[Index]).Free;
  inherited;
end;

destructor TRdfGraph.Destroy;
var
  I : Integer;
begin
  for I := Pred(FItems.Count) downto 0 do
    TRdfStatement(FItems[I]).Free;
  inherited;
end;

procedure TRdfGraph.DisableControls;
begin
  DoNotifyDisableControls;
end;

function TRdfGraph.DoCompare(const Object1,
  Object2: TObject): Integer;

  function CompareSubject(Stmt1, Stmt2: TRdfStatement): Integer;
  var
    S1, S2: string;
  begin
    S1 := EscapeStr(Stmt1.Subject_Reference);
    S2 := EscapeStr(Stmt2.Subject_Reference);
    Result := CompareStr(S1, S2);
  end;

  function ComparePredicate(Stmt1, Stmt2: TRdfStatement): Integer;
  begin
    Result := CompareStr(EscapeStr(Stmt1.Predicate_Reference),
                         EscapeStr(Stmt2.Predicate_Reference));
  end;

  function CompareObject(Stmt1, Stmt2: TRdfStatement): Integer;
  var
    S1, S2: string;
  begin
    with Stmt1 do
      case ReferenceTypeOf(Object_Reference) of
        rtGlobal, rtLocal:  S1 := EscapeStr(Object_Reference);
        rtNone:             S1 := '~' + EscapeStr(Object_Datatype) + '~' + Object_Language + '~' + Object_Lexical;
      end;
    with Stmt2 do
      case ReferenceTypeOf(Object_Reference) of
        rtGlobal, rtLocal:  S2 := EscapeStr(Object_Reference);
        rtNone:             S2 := '~' + EscapeStr(Object_Datatype) + '~' + Object_Language + '~' + Object_Lexical;
      end;
    Result := CompareStr(S1, S2);
  end;

begin
  Assert( (Object1 is TRdfStatement) and (Object2 is TRdfStatement) );

  case SortType of
    stSPO: begin
      Result := CompareSubject(TRdfStatement(Object1), TRdfStatement(Object2));
      if Result = 0 then begin
        Result := ComparePredicate(TRdfStatement(Object1), TRdfStatement(Object2));
        if Result = 0 then
          Result := CompareObject(TRdfStatement(Object1), TRdfStatement(Object2));
      end;
    end;

    stSOP: begin
      Result := CompareSubject(TRdfStatement(Object1), TRdfStatement(Object2));
      if Result = 0 then begin
        Result := CompareObject(TRdfStatement(Object1), TRdfStatement(Object2));
        if Result = 0 then
          Result := ComparePredicate(TRdfStatement(Object1), TRdfStatement(Object2));
      end;
    end;

    stPSO: begin
      Result := ComparePredicate(TRdfStatement(Object1), TRdfStatement(Object2));
      if Result = 0 then begin
        Result := CompareSubject(TRdfStatement(Object1), TRdfStatement(Object2));
        if Result = 0 then
          Result := CompareObject(TRdfStatement(Object1), TRdfStatement(Object2));
      end;
    end;

    stPOS: begin
      Result := ComparePredicate(TRdfStatement(Object1), TRdfStatement(Object2));
      if Result = 0 then begin
        Result := CompareObject(TRdfStatement(Object1), TRdfStatement(Object2));
        if Result = 0 then
          Result := CompareSubject(TRdfStatement(Object1), TRdfStatement(Object2));
      end;
    end;

    stOSP: begin
      Result := CompareObject(TRdfStatement(Object1), TRdfStatement(Object2));
      if Result = 0 then begin
        Result := CompareSubject(TRdfStatement(Object1), TRdfStatement(Object2));
        if Result = 0 then
          Result := ComparePredicate(TRdfStatement(Object1), TRdfStatement(Object2));
      end;
    end;

    stOPS: begin
      Result := CompareObject(TRdfStatement(Object1), TRdfStatement(Object2));
      if Result = 0 then begin
        Result := ComparePredicate(TRdfStatement(Object1), TRdfStatement(Object2));
        if Result = 0 then
          Result := CompareSubject(TRdfStatement(Object1), TRdfStatement(Object2));
      end;
    end;
  else
    Result := 0;
  end;
end;

function TRdfGraph.Edit(Index: Integer; RdfSubject,
  RdfPredicate, RdfObject, RdfObjectLexical, RdfObjectLanguage,
  RdfObjectDatatype: UTF8String): Integer;
begin
  if Sorted then begin
    Delete(Index);
    Result:= Add(
        RdfSubject,
        RdfPredicate,
        RdfObject,
        RdfObjectLexical,
        RdfObjectLanguage,
        RdfObjectDatatype
      );
  end else begin
    TRdfStatement(FItems[Index]).Free;
    FItems[Index] :=
      TRdfStatement.Create(
        Self,
        RdfSubject,
        RdfPredicate,
        RdfObject,
        RdfObjectLexical,
        RdfObjectLanguage,
        RdfObjectDatatype
      );
    Result := Index;
  end;
end;

procedure TRdfGraph.EnableControls;
begin
  DoNotifyEnableControls;
end;

function TRdfGraph.EscapeStr(S: string): string;
begin
  if Assigned(Aliases)
    then Result := Aliases.EscapeStr(S, aoPrefix)
    else Result := S;
end;

function TRdfGraph.Find(const Stmt: TRdfStatement;
  var Index: Integer): Boolean;
begin
  Result := inherited Find(Stmt, Index);
end;

function TRdfGraph.GetNewNodeID: UTF8String;
begin
  Inc(FLastNodeID);
  Result := '_:n' + IntToStr(FLastNodeID);
end;

function TRdfGraph.GetObject(Index: Integer): UTF8String;
begin
  Result := Statements[Index].Object_Reference;
end;

function TRdfGraph.GetObjectDatatype(Index: Integer): UTF8String;
begin
  Result := Statements[Index].Object_Datatype;
end;

function TRdfGraph.GetObjectLanguage(Index: Integer): UTF8String;
begin
  Result := Statements[Index].Object_Language;
end;

function TRdfGraph.GetObjectLexical(Index: Integer): UTF8String;
begin
  Result := Statements[Index].Object_Lexical;
end;

function TRdfGraph.GetPredicate(Index: Integer): UTF8String;
begin
  Result := Statements[Index].Predicate_Reference;
end;

function TRdfGraph.GetStatement(Index: Integer): TRdfStatement;
begin
  Result := inherited Objects[Index] as TRdfStatement;
end;

function TRdfGraph.GetStatementAttribute(Index: Integer;
  Kind: TRdfStmtAttr): string;
begin
  case Kind of
    saSubject:         Result := Subjects[Index];
    saPredicate:       Result := Predicates[Index];
    saObject:          Result := Objects[Index];
    saObjectLexical:   Result := ObjectLexicals[Index];
    saObjectLanguage:  Result := ObjectLanguages[Index];
    saObjectDatatype:  Result := ObjectDatatypes[Index];
  end;
end;

function TRdfGraph.GetSubject(Index: Integer): UTF8String;
begin
  Result := Statements[Index].Subject_Reference;
end;

function TRdfGraph.Insert(Index: Integer; RdfSubject,
  RdfPredicate, RdfObject, RdfObjectLexical, RdfObjectLanguage,
  RdfObjectDatatype: UTF8String): Boolean;
var
  NewStmt: TRdfStatement;
begin
  NewStmt := TRdfStatement.Create(Self,
                                  RdfSubject,
                                  RdfPredicate,
                                  RdfObject,
                                  RdfObjectLexical,
                                  RdfObjectLanguage,
                                  RdfObjectDatatype);
  try
    Result := inherited Insert(Index, newStmt);
  except
    NewStmt.Free;
    raise;
  end;
end;

procedure TRdfGraph.PutStatement(Index: Integer;
  const Value: TRdfStatement);
begin
  inherited Objects[Index] := Value;
end;

procedure TRdfGraph.SetSortType(Value: TRdfSortType);
begin
  CheckCanModify;
  if FSortType <> Value then begin
    if Value <> stNone then Sort;
    FSortType := Value;
  end;
end;

function TRdfGraph.UnescapeStr(S: string): string;
begin
  if Assigned(Aliases)
    then Result := Aliases.UnescapeStr(S, aoPrefix)
    else Result := S;
end;

{ TRdfCustomFilteredGraph }

function TRdfCustomFilteredGraph.Add(RdfSubject,
  RdfPredicate, RdfObject, RdfObjectLexical, RdfObjectLanguage,
  RdfObjectDatatype: UTF8String): Integer;
begin
  if MatchFilter(RdfSubject, RdfPredicate, RdfObject,
                 RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype)
    then Result := inherited Add(RdfSubject, RdfPredicate, RdfObject,
                                 RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype)
    else Result := -1;
end;

constructor TRdfCustomFilteredGraph.Create(AOwner: TComponent);
begin
  inherited;
  FFilter := TRdfStatementFilter.Create;
end;

destructor TRdfCustomFilteredGraph.Destroy;
begin
  FFilter.Free;
  FFilter := nil;
  inherited;
end;

procedure TRdfCustomFilteredGraph.DoOnFilterStatement(RdfSubject,
  RdfPredicate, RdfObject, RdfObjectLexical, RdfObjectLanguage,
  RdfObjectDatatype: UTF8String; out Accept: Boolean);
begin
  Accept := True;
  if Assigned(FOnFilterStatement)
    then FOnFilterStatement(Self, RdfSubject, RdfPredicate, RdfObject,
           RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype, Accept);
end;

function TRdfCustomFilteredGraph.Edit(Index: Integer; RdfSubject,
  RdfPredicate, RdfObject, RdfObjectLexical, RdfObjectLanguage,
  RdfObjectDatatype: UTF8String): Integer;
begin
  if MatchFilter(RdfSubject, RdfPredicate, RdfObject,
                 RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype)
    then Result := inherited Edit(Index, RdfSubject, RdfPredicate, RdfObject,
                                  RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype)
    else Result := -1;
end;

function TRdfCustomFilteredGraph.Insert(Index: Integer; RdfSubject,
  RdfPredicate, RdfObject, RdfObjectLexical, RdfObjectLanguage,
  RdfObjectDatatype: UTF8String): Boolean;
begin
  if MatchFilter(RdfSubject, RdfPredicate, RdfObject,
                 RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype)
    then Result := inherited Insert(Index, RdfSubject, RdfPredicate, RdfObject,
                                    RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype)
    else Result := False;
end;

function TRdfCustomFilteredGraph.MatchFilter(RdfSubject, RdfPredicate,
  RdfObject, RdfObjectLexical, RdfObjectLanguage,
  RdfObjectDatatype: UTF8String): Boolean;
begin
  if csDestroying in ComponentState then begin Result := False; Exit; end;
  Result := True;
  if AutoFilter then
    Result := Filter.Match(RdfSubject, RdfPredicate, RdfObject,
                           RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype);
  if Result then
    DoOnFilterStatement(RdfSubject, RdfPredicate, RdfObject,
                        RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype,
                        Result);
end;

procedure TRdfCustomFilteredGraph.SetAutoFilter(const Value: Boolean);
begin
  FAutoFilter := Value;
end;

procedure TRdfCustomFilteredGraph.SetFilter(const Value: TRdfStatementFilter);
begin
  FFilter.Assign(Value);
end;

{ TRdfFilteredGraph }

function TRdfFilteredGraph.Add(RdfSubject, RdfPredicate,
  RdfObject, RdfObjectLexical, RdfObjectLanguage,
  RdfObjectDatatype: UTF8String): Integer;
begin
  Result := inherited Add(RdfSubject, RdfPredicate, RdfObject,
                          RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype);
end;

procedure TRdfFilteredGraph.Clear;
begin
  inherited;
end;

procedure TRdfFilteredGraph.Delete(Index: Integer);
begin
  inherited;
end;

function TRdfFilteredGraph.Edit(Index: Integer; RdfSubject,
  RdfPredicate, RdfObject, RdfObjectLexical, RdfObjectLanguage,
  RdfObjectDatatype: UTF8String): Integer;
begin
  Result := inherited Edit(Index, RdfSubject, RdfPredicate, RdfObject,
                           RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype);
end;

function TRdfFilteredGraph.GetNewNodeID: UTF8String;
begin
  Result := inherited GetNewNodeID;
end;

function TRdfFilteredGraph.Insert(Index: Integer; RdfSubject,
  RdfPredicate, RdfObject, RdfObjectLexical, RdfObjectLanguage,
  RdfObjectDatatype: UTF8String): Boolean;
begin
  Result := inherited Insert(Index, RdfSubject, RdfPredicate, RdfObject,
                             RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype);
end;

{ TRdfCustomDetailGraph }

procedure TRdfCustomDetailGraph.ActiveChangedEventHandler(
  Sender: TObject);
begin
  UpdateView;
end;

constructor TRdfCustomDetailGraph.Create(AOwner: TComponent);
begin
  inherited;
  FGraphIterator := TRdfGraphIterator.Create(self);
  FAdapter := TRdfIteratorAdapter.Create;
  with FAdapter do begin
    SetIterator(FGraphIterator); // Setting the iterator must go before setting the event handlers.
    OnActiveChanged := ActiveChangedEventHandler;
    OnListModified := ListModifiedEventHandler;
  end;
  ReadOnly := True;
end;

destructor TRdfCustomDetailGraph.Destroy;
begin
  SetRdfGraph(nil);
  FAdapter.Free;
  FAdapter := nil;
  FGraphIterator.Free;
  inherited;
end;

procedure TRdfCustomDetailGraph.DoOnAfterUpdateView;
begin
  if Assigned(FOnAfterUpdateView) then FOnAfterUpdateView(Self);
end;

procedure TRdfCustomDetailGraph.DoOnBeforeUpdateView;
begin
  if Assigned(FOnBeforeUpdateView) then FOnBeforeUpdateView(Self);
end;

function TRdfCustomDetailGraph.GetRdfGraph: TRdfGraph;
begin
  Result := GraphIterator.RdfGraph;
end;

procedure TRdfCustomDetailGraph.ListModifiedEventHandler(Sender: TObject);
begin
  UpdateView;
end;

procedure TRdfCustomDetailGraph.SetRdfGraph(Value: TRdfGraph);
begin
  if not (csLoading in GraphIterator.ComponentState) then
    GraphIterator.RdfGraph := Value;
end;

procedure TRdfCustomDetailGraph.UpdateView;
begin
  if csDestroying in ComponentState then Exit;
  DoOnBeforeUpdateView;
  DisableControls;
  try
    ReadOnly := False;
    Clear;
    with GraphIterator do begin
      First;
      while not Offright do begin
        Add(RdfSubject, RdfPredicate, RdfObject,
            RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype);
        Next;
      end;
    end;
  finally
    ReadOnly := True;
    EnableControls;
  end;
  DoOnAfterUpdateView;
end;

{ TRdfCustomRestricterGraph }

constructor TRdfCustomRestricterGraph.Create(AOwner: TComponent);
begin
  inherited;
  FRestricterFilter := TRdfStatementFilter.Create;
  FSubjectRestricterAdapter := TUtilsStringIteratorAdapter.Create;
  with FSubjectRestricterAdapter do begin
    OnActiveChanged := ItemChangedEventHandler;
    OnItemChanged := ItemChangedEventHandler;
  end;
  FPredicateRestricterAdapter := TUtilsStringIteratorAdapter.Create;
  with FPredicateRestricterAdapter do begin
    OnActiveChanged := ItemChangedEventHandler;
    OnItemChanged := ItemChangedEventHandler;
  end;
  FObjectRestricterAdapter := TUtilsStringIteratorAdapter.Create;
  with FObjectRestricterAdapter do begin
    OnActiveChanged := ItemChangedEventHandler;
    OnItemChanged := ItemChangedEventHandler;
  end;
end;

destructor TRdfCustomRestricterGraph.Destroy;
begin
  FSubjectRestricterAdapter.Free;
  FSubjectRestricterAdapter := nil;
  FPredicateRestricterAdapter.Free;
  FPredicateRestricterAdapter := nil;
  FObjectRestricterAdapter.Free;
  FObjectRestricterAdapter := nil;
  FRestricterFilter.Free;
  FRestricterFilter := nil;
  inherited;
end;

function TRdfCustomRestricterGraph.GetObjectRestricter: TUtilsStringIterator;
begin
  Result := FObjectRestricterAdapter.Iterator;
end;

function TRdfCustomRestricterGraph.GetPredicateRestricter: TUtilsStringIterator;
begin
  Result := FPredicateRestricterAdapter.Iterator;
end;

function TRdfCustomRestricterGraph.GetSubjectRestricter: TUtilsStringIterator;
begin
  Result := FSubjectRestricterAdapter.Iterator;
end;

procedure TRdfCustomRestricterGraph.ItemChangedEventHandler(
  Sender: TObject);
begin
  UpdateView;
end;

function TRdfCustomRestricterGraph.MatchFilter(RdfSubject,
  RdfPredicate, RdfObject, RdfObjectLexical, RdfObjectLanguage,
  RdfObjectDatatype: UTF8String): Boolean;
begin
  if csDestroying in ComponentState then begin Result := False; Exit; end;
  Result := RestricterFilter.Match(RdfSubject, RdfPredicate, RdfObject,
                             RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype);

  if Result and AutoFilter then
    Result := Filter.Match(RdfSubject, RdfPredicate, RdfObject,
                           RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype);
  if Result then
    DoOnFilterStatement(RdfSubject, RdfPredicate, RdfObject,
                        RdfObjectLexical, RdfObjectLanguage, RdfObjectDatatype,
                        Result);
end;

procedure TRdfCustomRestricterGraph.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) then begin
    if Assigned(FSubjectRestricterAdapter) and (AComponent = SubjectRestricter)
      then SubjectRestricter := nil;
    if Assigned(FPredicateRestricterAdapter) and (AComponent = PredicateRestricter)
      then PredicateRestricter := nil;
    if Assigned(FObjectRestricterAdapter) and (AComponent = ObjectRestricter)
      then ObjectRestricter := nil;
  end;
end;

procedure TRdfCustomRestricterGraph.SetObjectRestricter(
  Value: TUtilsStringIterator);
begin
  if not (FObjectRestricterAdapter.IteratorFixed and (csLoading in ComponentState)) then
    FObjectRestricterAdapter.Iterator := Value;
  if Assigned(Value) then Value.FreeNotification(Self);
end;

procedure TRdfCustomRestricterGraph.SetPredicateRestricter(
  Value: TUtilsStringIterator);
begin
  if not (FPredicateRestricterAdapter.IteratorFixed and (csLoading in ComponentState)) then
    FPredicateRestricterAdapter.Iterator := Value;
  if Assigned(Value) then Value.FreeNotification(Self);
end;

procedure TRdfCustomRestricterGraph.SetSubjectRestricter(
  Value: TUtilsStringIterator);
begin
  if not (FSubjectRestricterAdapter.IteratorFixed and (csLoading in ComponentState)) then
    FSubjectRestricterAdapter.Iterator := Value;
  if Assigned(Value) then Value.FreeNotification(Self);
end;

procedure TRdfCustomRestricterGraph.UpdateView;
var
  S: string;
begin
  if csDestroying in ComponentState then Exit;
  with RestricterFilter do begin
    RejectAll := False;
    if Assigned(FSubjectRestricterAdapter.Iterator) then begin
      S := FSubjectRestricterAdapter.Iterator.LiteralValue;
      if IsNTripleUriRef(S) or IsNTripleNodeID(S) then begin
        SubjectQualifier := S;
      end else begin
        RejectAll := True;
        Exit;
      end;
    end else SubjectQualifier := '<*>';
    if Assigned(FPredicateRestricterAdapter.Iterator) then begin
      S := FPredicateRestricterAdapter.Iterator.LiteralValue;
      if IsNTripleUriRef(S) then begin
        PredicateQualifier := S;
      end else begin
        RejectAll := True;
        Exit;
      end;
    end else PredicateQualifier := '<*>';
    if Assigned(FObjectRestricterAdapter.Iterator) then begin
      S := FObjectRestricterAdapter.Iterator.LiteralValue;
      if IsNTripleUriRef(S) or IsNTripleNodeID(S) then begin
        ObjectQualifier := S;
      end else begin
        RejectAll := True;
        Exit;
      end;
    end else ObjectQualifier := '*';
  end;
  inherited;
end;

{ TRdfDetailGraph }

procedure TRdfDetailGraph.ForceUpdate;
begin
  UpdateView;
end;

{ TRdfGraphIterator }

function TRdfGraphIterator.GetCanModify: Boolean;
begin
  if Assigned(List)
    then Result := RdfGraph.CanModify and (UpdateCount = 0)
    else Result := False;
end;

function TRdfGraphIterator.GetRdfGraph: TRdfGraph;
begin
  Result := inherited GetList as TRdfGraph;
end;

function TRdfGraphIterator.GetRdfObject: UTF8String;
begin
  Result := FRdfObject;
end;

function TRdfGraphIterator.GetRdfObjectDatatype: UTF8String;
begin
  Result := FRdfObjectDatatype;
end;

function TRdfGraphIterator.GetRdfObjectLanguage: UTF8String;
begin
  Result := FRdfObjectLanguage;
end;

function TRdfGraphIterator.GetRdfObjectLexical: UTF8String;
begin
  Result := FRdfObjectLexical;
end;

function TRdfGraphIterator.GetRdfPredicate: UTF8String;
begin
  Result := FRdfPredicate;
end;

function TRdfGraphIterator.GetRdfSubject: UTF8String;
begin
  Result := FRdfSubject;
end;

procedure TRdfGraphIterator.SetRdfGraph(const Value: TRdfGraph);
begin
  inherited SetList(Value);
end;

procedure TRdfGraphIterator.UpdateIterator(const NewPos: Integer;
  const ItemChanged, ListModified: Boolean);
var
  NewCount: Integer;
begin
  if ItemChanged then begin
    if Assigned(RdfGraph) then begin
      if RdfGraph.Enabled
        then NewCount := RdfGraph.Count
        else NewCount := 0;
      if (NewPos = -1) or (NewPos = NewCount) then begin
        FRdfObject := '';
        FRdfObjectDatatype := '';
        FRdfObjectLanguage := '';
        FRdfObjectLexical := '';
        FRdfPredicate := '';
        FRdfSubject := '';
      end else begin
        FRdfObject := RdfGraph.Objects[NewPos];
        FRdfObjectDatatype := RdfGraph.ObjectDatatypes[NewPos];
        FRdfObjectLanguage := RdfGraph.ObjectLanguages[NewPos];
        FRdfObjectLexical := RdfGraph.ObjectLexicals[NewPos];
        FRdfPredicate := RdfGraph.Predicates[NewPos];
        FRdfSubject := RdfGraph.Subjects[NewPos];
      end;
    end else begin
      FRdfObject := '';
      FRdfObjectDatatype := '';
      FRdfObjectLanguage := '';
      FRdfObjectLexical := '';
      FRdfPredicate := '';
      FRdfSubject := '';
    end;
  end;

  inherited;
end;

{ TRdfRestricterGraph }

procedure TRdfRestricterGraph.ForceUpdate;
begin
  UpdateView;
end;

{ TRdfAliases }

constructor TRdfAliases.Create(AOwner: TComponent);
begin
  inherited;
  FPredicateSourceAdapter := TUtilsStringIteratorAdapter.Create;
  with FPredicateSourceAdapter do begin
    OnActiveChanged := GeneralEventHandler;
    OnItemChanged := GeneralEventHandler;
  end;
  FNameIterator := TRdfGraphIterator.Create(self);
  FNameAdapter := TRdfIteratorAdapter.Create;
  with FNameAdapter do begin
    SetIterator(FNameIterator); // Setting the iterator must go before setting the event handlers.
    OnActiveChanged := GeneralEventHandler;
    OnListModified := GeneralEventHandler;
  end;
  FAliasIterator := TRdfGraphIterator.Create(self);
  FAliasAdapter := TRdfIteratorAdapter.Create;
  with FAliasAdapter do begin
    SetIterator(FAliasIterator); // Setting the iterator must go before setting the event handlers.
    OnActiveChanged := GeneralEventHandler;
    OnListModified := GeneralEventHandler;
  end;
end;

destructor TRdfAliases.Destroy;
begin
  FPredicateSourceAdapter.Free;
  FPredicateSourceAdapter := nil;

  SetNameSource(nil);
  FNameAdapter.Free;
  FNameAdapter := nil;
  FNameIterator.Free;

  SetAliasSource(nil);
  FAliasAdapter.Free;
  FAliasAdapter := nil;
  FAliasIterator.Free;

  inherited;
end;

procedure TRdfAliases.DoOnAfterUpdate;
begin
  if Assigned(FOnAfterUpdate) then FOnAfterUpdate(Self);
end;

procedure TRdfAliases.DoOnBeforeUpdate;
begin
  if Assigned(FOnBeforeUpdate) then FOnBeforeUpdate(Self);
end;

function TRdfAliases.GetNameSource: TRdfGraph;
begin
  Result := FNameIterator.RdfGraph;
end;

function TRdfAliases.GetAliasSource: TRdfGraph;
begin
  Result := FAliasIterator.RdfGraph;
end;

function TRdfAliases.GetPredicateSource: TUtilsStringIterator;
begin
  Result := FPredicateSourceAdapter.Iterator;
end;

procedure TRdfAliases.GeneralEventHandler(Sender: TObject);
begin
  UpdateAliases;
end;

function TRdfAliases.IsPredicateURI(Value: UTF8String): Boolean;
begin
  if Length(Value) < 3 then begin
    Result := False;
    Exit;
  end;
  if (Value[1] <> '<') or (Value[Length(Value)] <> '>') then begin
    Result := False;
    Exit;
  end;
  Value := Copy(Value, 2, Length(Value) - 2); // Remove brackets
  Result := IsUriStr(Value);
end;

procedure TRdfAliases.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) then begin
    if Assigned(FPredicateSourceAdapter) and (AComponent = PredicateSource)
      then PredicateSource := nil;
  end;
end;

procedure TRdfAliases.SetNameSource(const Value: TRdfGraph);
begin
  if not (csLoading in FNameIterator.ComponentState) then
    FNameIterator.RdfGraph := Value;
end;

procedure TRdfAliases.SetAliasSource(const Value: TRdfGraph);
begin
  if not (csLoading in FAliasIterator.ComponentState) then
    FAliasIterator.RdfGraph := Value;
end;

procedure TRdfAliases.SetPredicateSource(
  const Value: TUtilsStringIterator);
begin
  if not (FPredicateSourceAdapter.IteratorFixed and (csLoading in ComponentState)) then
    FPredicateSourceAdapter.Iterator := Value;
  if Assigned(Value) then Value.FreeNotification(Self);
end;

procedure TRdfAliases.SetPredicateURI(const Value: UTF8String);
begin
  if (Value <> '') and not IsPredicateUri(Value) then
    AutoListError(SInvalidAbbr, -1);
  FPredicateURI := Value;
end;

procedure TRdfAliases.UpdateAliases;
var
  Predicate: UTF8String;
begin
  if csDestroying in ComponentState then Exit;
  DoOnBeforeUpdate;
  BeginUpdate;
  try
    Clear;

    if not ( Assigned(NameSource) and Assigned(AliasSource) ) then Exit;

    if Assigned(PredicateSource) then begin
      Predicate := PredicateSource.LiteralValue;
      if not IsPredicateUri(Predicate) then Exit;
    end else if PredicateURI = ''
      then Exit
      else Predicate := PredicateURI;

    with FNameIterator do begin
      First;
      while not Offright do begin

        if not HasKey(RdfSubject) then begin
          // xxx Generalize and move the search functionality to the TRdfGraphIterator class!
          with FAliasIterator do begin
            First;
            while not Offright do begin
              if (RdfSubject = FNameIterator.RdfSubject) and
                 (RdfPredicate = Predicate) then begin
                if RdfObject = ''
                  then Define(RdfSubject, RdfObjectLexical)
                  else Define(RdfSubject, RdfObject);
                Last;
              end;
              Next;
            end;
          end;
        end;

        Next;
      end;
    end;
  finally
    EndUpdate;
  end;
  DoOnAfterUpdate;
end;

end.
